﻿//An Idea to build a bus service
class BusServiceIdea extends Idea {
	roadSystem = null;
	townA = null
	townB = null
	constructor(roadSystem, townA, townB, data = null) {
		if (data) {
			this.roadSystem = data.misc[0]
			this.townA = data.misc[1]
			this.townB = data.misc[2]
			this.results = data.results
			this.plans = []
			this.plans.append(BusStationPlan(data.plans[0], null, null))
			this.plans.append(BusStationPlan(data.plans[1], null, null))
			this.plans.append(GaragePlan(data.plans[2], null, null))
			this.plans.append(RoadPlan(null, null, null, data.plans[3]))
			this.plans.append(RoadPlan(null, null, null, data.plans[4]))
			this.plans.append(RoadVehicleBuyPlan(null, null, data.plans[5]))
			this.plans.append(RoadPlan(null, null, null, data.plans[6]))
			this.plans.append(RoadPlan(null, null, null, data.plans[7]))
		} else {
			this.plans = []
			this.results = []
			LogWarning("Creating a Bus Service Idea: " + AITown.GetName(townA) + " to " + AITown.GetName(townB))
			this.roadSystem = roadSystem
			this.townA = townA
			this.townB = townB
			AIRoad.SetCurrentRoadType(roadSystem[0])
			local plan
			plan = BusStationPlan(null, AITown.GetLocation(townA), roadSystem, true)
			this.plans.append(plan)
			plan = BusStationPlan(null, AITown.GetLocation(townB), roadSystem, false)
			this.plans.append(plan)
			if (!this.plans[0].location || !this.plans[1].location) {
				this.plans = []
				return
			}
			plan = GaragePlan(null, this.plans[0].location, roadSystem)
			this.plans.append(plan)
			if (this.plans[0].frontTile && this.plans[1].frontTile && this.plans[2].location && plan.Build(true)) {
				plan = RoadPlan(roadSystem, this.plans[0].getData(), this.plans[1].getData())
				this.plans.append(plan)
				if (plan.path1) {
					plan = RoadPlan(roadSystem, this.plans[0].getData(), this.plans[2].getData())
					this.plans.append(plan)
					if (plan.path1) {
						plan = RoadVehicleBuyPlan(roadSystem, this.plans[2])
						this.plans.append(plan)
						plan = RoadPlan(roadSystem, this.plans[0].getData(), this.plans[0].getData())
						this.plans.append(plan)
						plan = RoadPlan(roadSystem, this.plans[1].getData(), this.plans[1].getData())
						this.plans.append(plan)
						if (!this.plans[6].path1 || !this.plans[7].path1) {
							this.plans = []
							LogWarning("Could not plan to build roads within the city")
						}
					} else {
						this.plans = []
						LogWarning("Could not plan to build roads towards the garage")
					}
				 } else {
					this.plans = []
					LogWarning("Could not plan to build roads")
				}
			} else {
				LogWarning("Could not plan both road stations or the garage")
				this.plans = []
			}
		}
	}
	
	function getType() {
		return "BusServiceIdea"
	}
	
	function getMiscData() {
		return [roadSystem, townA, townB]
	}
	
	function Execute(fromSave = false) {
		AIRoad.SetCurrentRoadType(roadSystem[0])
		if (!Perform(fromSave)) {
			LogError("Idea could not be executed")
			return false
		}
		local breakdownsEnabled = (AIGameSettings.GetValue("difficulty.vehicle_breakdowns") >= 1)
		local vehicle = this.results[5]
		//First order is go to garage where it was built. Is used for cloning the bus later
		if (breakdownsEnabled) {
			AIOrder.AppendOrder(vehicle, this.results[2].location, AIOrder.OF_SERVICE_IF_NEEDED);
		} else {
			AIOrder.AppendOrder(vehicle, this.results[2].location, AIOrder.OF_NONE);
		}
		//Go to the first station
		AIOrder.AppendOrder(vehicle, this.results[0]["location"], AIOrder.OF_NONE);
		//Don't go to the garage
		AIOrder.InsertConditionalOrder(vehicle, 0, 1)
		AIOrder.SetOrderCondition(vehicle, 0, AIOrder.OC_RELIABILITY)
		AIOrder.SetOrderCompareFunction(vehicle, 0, AIOrder.CF_MORE_THAN)
		if (breakdownsEnabled) {
			AIOrder.SetOrderCompareValue(vehicle, 0, 70)
		} else {
			AIOrder.SetOrderCompareValue(vehicle, 0, 0)
		}
		
		// Do a round trip if the vehicle is exactly 1% full, then we have a tourist!
		AIOrder.AppendConditionalOrder(vehicle, 2);
		AIOrder.SetOrderCondition(vehicle, 3, AIOrder.OC_LOAD_PERCENTAGE);
		AIOrder.SetOrderCompareFunction(vehicle, 3, AIOrder.CF_LESS_THAN)
		AIOrder.SetOrderCompareValue(vehicle, 10, 1);
		// Go to the second bus stop
		AIOrder.AppendOrder(vehicle, this.results[1]["location"], AIOrder.OF_NONE);
		
		local group = AIGroup.CreateGroup(AIVehicle.GetVehicleType(vehicle))
		AIGroup.MoveVehicle(group, vehicle)
		
		AIVehicle.StartStopVehicle(vehicle);
		return {roadSystem = roadSystem, stops = [this.results[0], this.results[1]], trains = [this.results[3]]}
	}
}
